package bvh.jmd;

import java.io.*;
import java.util.*;
import java.util.Random;
import java.util.zip.*;
import java.awt.geom.Point2D;
import java.awt.Color;
import java.text.*;

import robocode.*;
import robocode.util.Utils;
//import jam.tools.*;

/*
import robocode.robocodeGL.*;
import robocode.robocodeGL.system.*;
*/

/**
 * Jormungand - a robot by Bart van Hest
 *
 *   "Jormungand, the World Snake, was the second eldest child between Loki the 
 *    Trickster and the giantess Angrboda. Kidnapped by the gods, he was cast 
 *    into the waters that surround Midgard. There he laid until he grew so 
 *    large that he ate large whales and was able to wrap himself around the 
 *    world and bite his own tail."
 *
 * Wijzigingshistorie v0.35 (18-05-2004):
 *  - kogelkracht niet langer gebaseerd op constante overbruggingstijd, maar
 *    meer random om het de wavesurfers lastiger te maken.
 *
 * Wijzigingshistorie v0.34 (18-05-2004):
 *  - weegfactoren in potentiaalmethode aangepast.
 *
 * Wijzigingshistorie v0.33 (18-05-2004):
 *  - GF-kanon aangepast waarbij altijd waves worden afgevuurd, niet alleen 
 *    wanneer ook daadwerkelijk een kogel wordt afgevuurd.
 *
 * Wijzigingshistorie v0.32 (18-05-2004):
 *  - bug in GF-kanon gecorrigeerd waardoor pas na ronde 4 werd geschoten!.
 *
 * Wijzigingshistorie v0.3 (18-05-2004):
 *  - bug in GF-kanon gecorrigeerd waardoor GF-statistieken niet werden bijgewerkt.
 *
 * Wijzigingshistorie v0.2 (29-02-2004):
 *  - verbeterde beweging gebaseerd op Fenrir 0.322
 *
 * Wijzigingshistorie v0.1 (27-02-2004):
 *  - gebaseerd op Fenrir 0.300
 *  - GF-kanon (in plaats van PM-kanon)
 *
 */

public class Jormungand extends AdvancedRobot implements Constanten {
/**
* run: Jormungand's hoofdroutine:
*/
   public void run() {

      // initieer tank kleuren:
      setColors(Color.black, Color.black, Color.pink);

// initialisatie tank:
      setAdjustGunForRobotTurn(adjustGun);
      setAdjustRadarForGunTurn(adjustRadar);

// initieer slagveld (kan alleen in run-method):
      slagveld      = new Slagveld(this);

// bepaal type gevecht( o.b.v. het aantal tegenstanders):
      if ( getOthers() > 1 ) meleeModus = true;
      else                   meleeModus = false;

/*
log("Start initialisation Narcissus");
      narc.newRound(this);
log("ended initialisation Narcissus");
*/
// start radar:
      setTurnRadarRightRadians( Double.POSITIVE_INFINITY );

// levensloop:
      while(true) {
         stuurRadar();  // stuur radar aan:
         navigatie();   // verplaats robot:
         execute();
      } // einde levensloop
   } // einde run-method

/**********************************************************************
** Methods voor event-afhandeling:
**********************************************************************/
/**
 * onScannedRobot: What to do when you see another robot
 * let op: van alle scanned bots tijdens een radar-sweep wordt de dichtstbijzijnde bot doorgegegeven!
 */
   public void onScannedRobot(ScannedRobotEvent evt) {
//      narc.onScannedRobot(evt);

      Doel   opponent;
// doel nieuw opvoeren in lijst of opzoeken en bijwerken:
      opponent = memoreerDoel(evt);

// oportunistisch handelen: alleen doel selecteren als energie laag is en kans op vernietiging groot (Loki trekje :-o )
//                          of als doel dichterbij is
//                          of als ik geen doel in het leven heb:
      if (doel.locked     == false           // selecteer gunstig doel indien nog geen ander doel gevonden
      || evt.getName()    == doel.naam       // update doel indien al eerder gevonden
      || evt.getDistance() < 200D            // selecteer gunstiger doel
      || evt.getEnergy()   < 17D             // selecteer gunstiger doel
      || (doel.afstand     > evt.getDistance()
        && doel.e          > evt.getEnergy() )
         )
         doel = opponent;  // leg informatie over doel vast:

      // kanonsturing:
      stuurKanon(new Point2D.Double(getX(), getY()), doel.positie, evt);
   }

/**
* onHitByBullet: What to do when you're hit by a bullet.
* This method will be called when your robot is hit by a bullet.
*/
   public void onHitByBullet(HitByBulletEvent evt) {
      if (!meleeModus) {
         if (Math.random()>0.5)
            stop();
         else
         {  setMaxVelocity(8D);
            bewegingsrichtingOmkeren();
            setAhead( (-2D*voorwaards+1D) * getDistanceRemaining() );
         }
      }
/*
      Doel   opponent;
      opponent = memoreerDoel(evt.getName());
      opponent.setInfo(evt, this);  // leg informatie over doel vast:
      opponent.setTreffers(evt); // update treffers door tegenstander op Jormungand
*/
   }

/**
* onBulletHit: What to do when your bullet hits a target.
* This method will be called when one of your bullets hits another robot.
*/
   public void onBulletHit(BulletHitEvent evt) {
/*
      Doel   opponent;
      opponent = memoreerDoel(evt.getName());
      opponent.setInfo(evt, this);  // leg informatie over doel vast:
      opponent.setTreffersBvh(evt);  // update mijn treffers op doel
*/
   }

/**
* onHitRobot: What to do when you're hit by a robot.
*/
   public void onHitRobot(HitRobotEvent evt) {
/*
      Doel   opponent;
      opponent = memoreerDoel(evt.getName());
      opponent.setInfo(evt, this);  // leg informatie over doel vast:
*/
   }

/**
* onRobotDeath: indien doel vernietigd dan dit 'vrijgeven'.
*/
   public void onRobotDeath(RobotDeathEvent evt) {
log("onRobotDeath: dood      ="+evt.getName());
log("onRobotDeath: doel.naam ="+doel.naam);
      // indien huidige doel is vernietigd, dan mijn bot initieren:
      if  (evt.getName().equals(doel.naam) ) {
         doel.locked = false;
log("onRobotDeath: doel.locked is false");
      }

      Doel   opponent;
      opponent = memoreerDoel(evt.getName());
      opponent.locked = false;

      stop();
   }

/**
* onWin: Niet stil blijven zitten om laatste op mij afgevuurde kogels te ontwijken.
* (zie UltimateDeathToNanoBots)
*/
   public void onWin(WinEvent e) {
      afronden();
      log("GEWONNEN: " + beurtOvergeslagen + " keren beurt overgeslagen.");
   }


/**
* onDeath: toon statistiek
*/
   public void onDeath(DeathEvent e) {
      afronden();
      log("....DOOD: " + beurtOvergeslagen + " keren beurt overgeslagen.");
   }

/**
* onSkippedTurn:  telt aantal keren dat beurt wordt overgeslagen;
*/
   public void onSkippedTurn(SkippedTurnEvent e) {
      beurtOvergeslagen += 1;
   }

/**********************************************************************
** Methods voor besturing tank
**********************************************************************/
/**
* navigatie(): Beweging o.b.v. potentiaalveld waarbij minder gunstige
* posities een hogere potentiaal krijgen en bot richting laagste potentiaal
* beweegd.
*/
   private void navigatie() {
      huidigePositie.setLocation( getX(), getY() );  // werken met oude, huidige en nieuwe positie !!!

      vGem= 0.857*vGem + 0.143*getVelocity();

      if ( Math.abs(getDistanceRemaining()) < 10D ) { // mogelijk verbetering door uitwijken te verwijderen!
         nieuwePositie = potentiaalMode(huidigePositie);

         if ( nieuwePositie.distance(huidigePositie) < 15 ) {
            nieuwePositie = SandboxMiniMode(huidigePositie, doel);
         }

         nieuweRichting   = Utils.normalRelativeAngle(HulpRoutines.bepaalRichting(nieuwePositie, huidigePositie) - getHeadingRadians());
         double draaiHoek = Math.atan(Math.tan(nieuweRichting));

// draai en verplaats robot:
         setTurnRightRadians(draaiHoek);
         setAhead(huidigePositie.distance(nieuwePositie) * (nieuweRichting == draaiHoek ? 1 : -1));

//         destPoint.setPosition(nieuwePositie.x, nieuwePositie.y);
//         destLabel.setString((int)nieuwePositie.x + ", " + (int)nieuwePositie.y);

         oudePositie.setLocation(huidigePositie.x, huidigePositie.y);
      }

// eventueel snelheid minderen om sneller te kunnen draaien:
      if ( Math.abs(getTurnRemainingRadians()) > KWARTPI ) {
         setMaxVelocity( (HALFPI-Math.abs(getTurnRemainingRadians()))/7.5D );
      }
      else
         setMaxVelocity(MAXIMUM_SNELHEID);
   }

   private Point2D.Double SandboxMiniMode(Point2D.Double positie, Doel d) {
      double   richting;
      richting = -(d.afstand-300D)/200D*KWARTPI;
      richting = (HALFPI + richting);
      return( bepaalXY(positie, richting) );
   }

   private Point2D.Double potentiaalMode(Point2D.Double positie) {
      Doel     o;
      double   richting;
      double factor;
      double[] potentiaal  = {0D, 0D, 0D, 0D, 0D, 0D, 0D, 0D};
      int      kwadrant, t, l;
      kwadrant = l = 8;

      // betrek wanden in potentiaal:
      if ( positie.x < WANDAFSTAND ) {
         potentiaal[5] += 7;
         potentiaal[6] += 9;
         potentiaal[7] += 7;
      }
      if ( positie.x > (getBattleFieldWidth()  - WANDAFSTAND) ) {
         potentiaal[1] += 7;
         potentiaal[2] += 9;
         potentiaal[3] += 7;
      }
      if ( positie.y < WANDAFSTAND ) {
         potentiaal[3] += 7;
         potentiaal[4] += 9;
         potentiaal[5] += 5;
      }
      if ( positie.y > (getBattleFieldHeight()  - WANDAFSTAND) ) {
         potentiaal[0] += 7;
         potentiaal[1] += 9;
         potentiaal[7] += 7;
      }

      // bepaal potentiaal door te sommeren over tegenstanders:
      for (Enumeration e = doelen.elements(); e.hasMoreElements(); ) {
         o               = (Doel)e.nextElement();
         if (o.locked) {
            factor          = Math.max(Math.min(Math.round(600D/o.afstand), 2D), 0.5);
            if (!meleeModus)
               factor         *= Math.max(Math.min(Math.round(o.e/50D), 2D), 0.5);
            kwadrant        = (int)(Math.round(l+l*o.richting/TWEEPI))%l;
            if (meleeModus) {
               potentiaal[ kwadrant]            += factor*3.0;
               potentiaal[(kwadrant+ l/2)%l]    += factor*2.0;
            }
            else {
//               potentiaal[ kwadrant]            += factor*(2D+random.nextInt(3));    // niet naar tegenstander toe bewegen
//               potentiaal[(kwadrant+ l/2)%l]    += factor*(1D+random.nextInt(3));         // niet van tegenstander af bewegen
               potentiaal[ kwadrant]            += factor*3.0;
               potentiaal[(kwadrant+ l/2)%l]    += factor*2.0;
            }
            potentiaal[(kwadrant + 1)%l]     += factor*1.0;                       // en kans op beweging schuin vanaf/naar tegenstander ook verkleinen
            potentiaal[(kwadrant - 1 + l)%l] += factor*1.0;
            potentiaal[(kwadrant + 3)%l]     += factor*0.5;
            potentiaal[(kwadrant - 3 + l)%l] += factor*0.5;
         } // controle op ge'locked' zijn van tegenstander
      }

// uitbreiden potentiaal door laatste eigen bewegingsrichting te betrekken (bewegingsrichting
// omkeren wordt ontmoedigd)
// proefondervindelijk aangetoond dat dit verbetering geeft in 1-vs-1,
// verslechtering bij melee-gevechten:

      if (getOthers() == 1) {
         // afstoting van vorige positie:

         kwadrant        = (int)(Math.round(l+l*HulpRoutines.bepaalRichting(oudePositie, positie)/TWEEPI))%l;
         potentiaal[kwadrant]  += 0.75;  // bewegingsrichting niet omkeren

         // aantrekking naar middelpunt slaveld:
         kwadrant        = (int)(Math.round(l+l*HulpRoutines.bepaalRichting(slagveld.middelpunt, positie)/TWEEPI))%l;
         potentiaal[kwadrant]  -= 0.5;   // naar middelpunt slagveld
      }


// uitbreiden potentiaal door nabije kogels te betrekken;
// indien kogel is gepasseerd, dan kogel verwijderen uit tabel:
      Kogel            kogel;
      double           kogelStraal, testAfstand;
      long             nu = getTime();

      for (Enumeration k = kogels.elements(); k.hasMoreElements(); ) {
// bepaal positie kogel t.o.v. eigen bot
         kogel           = (Kogel)k.nextElement();
         kogelStraal     = kogel.getAfgelegdeWeg(nu);
         testAfstand     = positie.distance( kogel.getScanPositie() ) - kogelStraal;

// nabije kogels met energie >= 1 meenemen in uitwijk-beweging:
         if ( (testAfstand <= 76D && testAfstand >= 0D) && kogel.v <= 17D ) {
            // kwadrant bepalen door huidige kogelpositie te schatten obv de
            // aanname dat deze recht op mijn bot is afgeschoten
            kwadrant        = (int)(Math.round(l+l*HulpRoutines.bepaalRichting( kogel.getPositie(nu), positie )/TWEEPI))%l;

            if (meleeModus) {
               potentiaal[kwadrant]             += 0.50;
               potentiaal[(kwadrant+ l/2)%l]    += 0.25;
               potentiaal[(kwadrant + 1)%l]     += 0.12;
               potentiaal[(kwadrant - 1 + l)%l] += 0.12;
            }
            else {
               potentiaal[kwadrant]             += 0.50;
               potentiaal[(kwadrant + 1)%l]     += 0.25;
               potentiaal[(kwadrant - 1 + l)%l] += 0.25;
               potentiaal[(kwadrant + 2)%l]     += 0.12;
               potentiaal[(kwadrant - 2 + l)%l] += 0.12;
            }

            if ( kogel.v <= 14D ) {
               uitwijken  = true;
            }
         }

         if ( testAfstand < 0D ) {
            kogels.remove(kogel.naam);
         }
      }


      // bepaal gunstigste richting:
      double minimum = 1D/0D;
      for (int i=0; i < l ; i++) {

//         potPoint[i].setPosition( positie.x+36*Math.sin(i*KWARTPI), positie.y+36*Math.cos(i*KWARTPI) );
//         potPoint[i].setSize( Math.round(potentiaal[i]+0.5) );

         if (    potentiaal[i] < minimum
            || ( potentiaal[i]== minimum && Math.random() > 0.25 )
            )
            minimum  = potentiaal[kwadrant = i];
      }
      richting = kwadrant*KWARTPI+ random.nextGaussian()*EENACHTSTEPI;
      return( bepaalXY(positie, richting) );
   }

   private Point2D.Double bepaalXY(Point2D.Double positie, double richting) {
      double r;
      // inbouwen willekeur in bewegingsafstand en tijdstip waarop volgende nieuwe
      // positie wordt bepaald.
      r = 1D + 3D*random.nextGaussian();

      // bepaal nieuwe x,y positie o.b.v. huidige positie, nieuwe hoek t.o.v. doel
      // en loopafstand die afhankelijk is van huidige afstand tot doel en hierboven
      // bepaalde random factor.
      nieuwePositie.x = getX() + r * STANDAARDLOOPAFSTAND * Math.sin(richting);
      nieuwePositie.y = getY() + r * STANDAARDLOOPAFSTAND * Math.cos(richting);
      // ontwijk wanden door min./max. waarden te bepalen voor nieuwe x,y positie.
      // overleving verbeterd door positie te spiegelen t.o.v. minimum wandafstand!
      nieuwePositie.x = Math.min(getBattleFieldWidth()  - WANDAFSTAND, Math.max(WANDAFSTAND, nieuwePositie.x));// * 2D - nieuwePositie.x;
      nieuwePositie.y = Math.min(getBattleFieldHeight() - WANDAFSTAND, Math.max(WANDAFSTAND, nieuwePositie.y));// * 2D - nieuwePositie.y;

      return( nieuwePositie );
   }

   private void bewegingsrichtingOmkeren() {
      if      (nieuweRichting < 0 ) nieuweRichting += PI;  // 180 graden bij nieuwe richting optellen
      else if (nieuweRichting > 0 ) nieuweRichting -= PI;  // 180 graden bij nieuwe richting aftrekken

      if (voorwaards == 1) voorwaards = 0;  // en bewegingsrichting omkeren
      else                 voorwaards = 1;
   }

/**********************************************************************
** Methods voor besturing kanon
**********************************************************************/
/**
* Vuurkracht afhankelijk van de doelafstand en de (voorspelde) hoek tussen kanon en doel
*
* v0.2 (28-08-2002)
*/
   private void stuurKanon( Point2D.Double botPositie
                          , Point2D.Double doelPositie
                          , ScannedRobotEvent evt){

      double kogelSnelheid, kanonVuurkracht, doelHoek, maxRichting;

      if ( doel.locked ) { // alleen kanon sturen als er op iets te schieten valt

         if ( doel.e <= 16D ) {
// energie tegenstander zodanig dat deze met n schot kan worden afgemaakt.
// bepaal hiervoor benodigde minimum vuurkracht:
            kanonVuurkracht = berekenMinimumVuurkracht( doel.e );
         }
         else {
// randomised kanonVuurkracht om het de wavesurfers/dodgers lastiger te maken
            int afstandIndex = (int)(doel.afstand/100D);
            kanonVuurkracht = 3D + 0.4 * afstandIndex * Math.random();
            kanonVuurkracht = Math.max( Math.min( kanonVuurkracht, 3D ), 0.1);
         }
         kogelSnelheid   = berekenKogelSnelheid( kanonVuurkracht );

         MicroGolf mg;
         mg = new MicroGolf();
         int afstandIndex   = Math.min( (int)Math.round(Math.sqrt(doel.afstand/25D)), 5 );

         // 0: doelPositie rond middelpunt slagveld
         // 1: doelPositie in tussengebied (meest waarschijnlijk)
         // 2: doelPositie in een van de hoeken
         int positieIndex   = slagveld.minimumCirkel.contains(doelPositie) ? 0 : (slagveld.maximumCirkel.contains(doelPositie) ? 1 : 2);
         int acclIndex      = (int)Math.abs(Math.round(Math.abs(doel.v_o)-Math.abs(doel.v)));
         if (acclIndex != 0)
            acclIndex = (acclIndex < 0)? 1 : 2;
         int snelheidsIndex = (int)(Math.abs(doel.v)/3);

         mg.startPositie     = botPositie;
         mg.richting         = doel.richting;
         mg.afstand          = 0;
         mg.kogelSnelheid    = kogelSnelheid;
         mg.cirkelRichting   = doel.v*Math.sin(doel.bewegingsrichting - doel.richting) > 0 ? .7D/GF_MIDDEN : -.7D/GF_MIDDEN;
// segmentatie o.b.v. 1:afstand
//                  , 2:huidige doel positie in middelpunt, neutraal of in hoeken.
//                  , 3:VoorspeldePositieBinnenSlagveld
//                  , 4:DoelSnelheid
         try {
            mg.stats            = doel.guessFactors[afstandIndex]
                                                   [positieIndex]
                                                   [acclIndex]
                                                   [snelheidsIndex];
         } catch (java.lang.ArrayIndexOutOfBoundsException e){
log("afstandIndex ="+ afstandIndex + " op (" +AFSTANDSEGMENTEN +") segmenten");
log("positieIndex ="+ positieIndex + " op (" +POSITIESEGMENTEN +") segmenten");
log("acclIndex    ="+ acclIndex + " op (" + VERSNELLINGSSEGMENTEN+") segmenten");
log("enemyVelocity="+ snelheidsIndex + " op (" +SNELHEIDSSEGMENTEN +") segmenten");
         }

         if ( doel.v == 0 && doel.e == 0 )
            doelHoek = doel.richting;
         else 
         {
            int bestGF = GF_MIDDEN;
            for (int gf = (2*GF_MIDDEN); gf >= 0 && doel.e > 0; gf--) //Saves one byte compared to going up, weird
               if (mg.stats[gf] > mg.stats[bestGF])
                  bestGF = gf;

            doelHoek = doel.richting + mg.cirkelRichting*(bestGF-GF_MIDDEN);
         }

         setTurnGunRightRadians(Utils.normalRelativeAngle( doelHoek - getGunHeadingRadians() ));

         if (getEnergy()   > kanonVuurkracht
         && (meleeModus || getGunTurnRemainingRadians() < Math.atan2(18D,doel.afstand))
            )
         {
            if ( setFireBullet( kanonVuurkracht) != null )
               doel.setSchotenBvh();
            addCustomEvent(mg);
         } // einde controle op uitlijnen kanon

      } // einde controle op aanwezigheid doel
   } // einde method stuurKannon()

/**
* berekenKogelSnelheid: berekend de snelheid van de kogel op basis
*    van de kracht van de kogel.
*
* v0.1 (27-12-2002)
*/
   private double berekenKogelSnelheid(double kanonVuurkracht) {
      return(20 - 3 * kanonVuurkracht);
   }
/**
* berekenMinimumKogelVuurkracht: berekend de benodigde vuurkracht van de kogel
*    om, op  basis van de energie van het doel, het doel met n kogel
*    uit te schakelen.
*
* v0.1 (27-12-2002)
*/
   private double berekenMinimumVuurkracht(double energie) {
      double vuurkracht;
      if (energie <= 12D ) vuurkracht = energie/4D;
      else                 vuurkracht = (energie+2D)/6D;

      vuurkracht    = vuurkracht + 0.1; // en beetje meer energie om doel te vernietigen...

      return( vuurkracht );
   }

/**
* berekenVuurkracht: berekend de vuurkracht van de kogel obv gegeven snelheid.
*
* v0.1 (22-10-2003)
*/
   private double berekenVuurkracht(double kogelSnelheid) {
      return( (20D - kogelSnelheid)/3D );
   }

/**
* berekenSnelheidBijConstanteOverbruggingstijd: berekend de sneleid van de kogel
*    obv gegeven afstand tot doel, zodanig dat de overbruggingstijd vrijwel
*    constant blijft voor iedere afgevuurde kogel. Het idee is dat dit resulteerd
*    in een nauwkeuriger patroonherkenning van de beweging van de tegenstander.
*
* v0.1 (22-10-2003)
*/
   private double berekenSnelheidBijConstanteOverbruggingstijd( double afstand ) {
      return( Math.min( Math.max( (afstand / 29.5D), maxKogelsnelheid), minKogelsnelheid ) );
   }
/**********************************************************************
** Methods voor besturing radar
**********************************************************************/
/**
* stuurRadar(): Radarsturing in melee gevechten waarbij zo groot mogelijk gebied wordt bestreken:
*
* v0.4 (10-04-2003): 1-vs-1 mode toegevoegd zodat een constanter scan-tussentijd wordt bereikt.
* v0.3 (25-12-2002): radar draait door tot alle bots in melee zijn gescanned waarna de
*                    draairichting omkeerd. Op deze manier hoeft niet perse 360 graden
*                    te worden gescanned en worden toch alle bots bijgehouden.
* v0.2 (09-11-2002): bestreken gebied wordt gebaseerd op de max. en min. hoek waarbinnen zich
*                    alle tegenstanders bevinden.
* v0.1 (10-09-2002): volledige 360 graden wordt bestreken.
*/
   private void stuurRadar() {
      if (getOthers() == 1 && doel.locked) {
         setTurnRadarRightRadians(3D*HulpRoutines.normaliseerRichting( doel.richting - getRadarHeadingRadians() ));
      }
      else
      // else (melee-mode) laat de radar draaien in dezelfde richting om een
      // gelijkmatige sample-rate te verkrijgen voor de patroon herkenning.
         setTurnRadarRightRadians( Double.POSITIVE_INFINITY );
   }

/**********************************************************************
** Methods voor doel beheersing
**********************************************************************/
/**
* methoden om doelen te bewerken:
*/
   private Doel memoreerDoel(ScannedRobotEvent evt) {
//log("memoreerDoel(ScannedRobotEvent evt)");
      String naam = evt.getName();
      Doel   opponent;

      if( doelen.containsKey(naam) ) {
//log("memoreerDoel(): doel "+naam+" bestaat al");
         opponent = (Doel)doelen.get(naam);
      }
      else {
log("memoreerDoel(): doel "+naam+" bestaat niet: inlezen data?");
         opponent = new Doel();
         doelen.put(naam, opponent);
         inlezenDoelData(naam, opponent);
      }
      opponent.setInfo(evt, this);  // leg informatie over doel vast:

      return opponent;
   }

   private Doel memoreerDoel(String naam) {
log("memoreerDoel(String naam)");
      Doel opponent;
// doel nieuw opvoeren in lijst of opzoeken en bijwerken:
      if( doelen.containsKey(naam) ) {
         opponent = (Doel)doelen.get(naam);
      } else {
         opponent = new Doel();
         doelen.put(naam, opponent);
         inlezenDoelData(naam, opponent);
      }
      return opponent;
   }

   private void verwijderDoel(String naam) {
      doelen.remove(naam);
   }

/**
* geefDichtstbijzijndeKogel(String naam): haalt de meest nabije kogel uit
*    de lijst met kogels.
*
* v0.2 (28-01-2004)
*  - verbetering bepaling afstand tot eigen bot door kogelstraal te nemen en
*    geschatte kogelrichting te negeren.
*
* v0.1 (23-06-2003)
*/
   private Kogel getDichtstbijzijndeKogel()
   {
      long       nu = getTime();
      double      a,
                  kortsteAfstand = 1D/0D;
      String      n = null;
      Kogel       k;

      for (Enumeration ko = kogels.elements(); ko.hasMoreElements(); )
      {
         k = (Kogel)ko.nextElement();
         a = Math.abs( huidigePositie.distance(k.getScanPositie()) - k.getAfgelegdeWeg(nu) );
         if (a < kortsteAfstand)
         {
            kortsteAfstand = a;
            n              = k.naam;
         }
      }

      if (n != null) k = (Kogel)kogels.get(n);
      else           k = null;

      return (k);
   }

/**********************************************************************
** methods voor afronden gevecht:
**********************************************************************/
   private void afronden() {
      saveSatistiek(); // per ronde statistiek tonen en naar file schrijven voor studie

      // na laatste ronde aan einde van gevecht de opgebouwde data-arrays opslaan naar file
      if (getRoundNum() == getNumRounds()-1)
      {
//         narc.saveData();

         Doel o;
         for (Enumeration en = doelen.elements(); en.hasMoreElements(); )
         {
            o = (Doel)en.nextElement();
            opslaanDoelData( o );
         }
      }
   }

/**********************************************************************
** methods voor weergeven/opslaan statistiek m.b.t. hits/missers/matches
** en gemiddelde afstand voor elk doel.
**********************************************************************/
   private void saveSatistiek(){
      String tekst;
      Doel d;
      for (Enumeration en = doelen.elements(); en.hasMoreElements(); ) {
         d = (Doel)en.nextElement();
log("saveStatistiek(doel.naam = "+d.naam+" )");
         try {
            log(">>**************************************************<<");
            log("Statistiek bot "+ d.naam);
            log(">>**************************************************<<");
            log("I fire bullet amount       : "+ d.getSchotenBvh() );
            log("My hit bullet amount       : "+ d.getTreffersBvh() );
            log("                 Hit ratio : "+100.0d*d.getTreffersBvh()/d.getSchotenBvh()+"%");
            log("+-------------------------------------+");
            log("Enemy fire bullet amount   : "+ d.getSchoten() );
            log("Enemy  hit bullet amount   : "+ d.getTreffers() );
            log("                 Hit ratio : "+100.0d*d.getTreffers()/d.getSchoten()+"%");
            log("+-------------------------------------+");
            log("Gemiddelde afstand         : "+d.afstandGem);
/*
            PrintStream w = new PrintStream(new RobocodeFileOutputStream(getDataFile(d.naam+".txt")));
            w.println(">>**************************************************<<");
            w.println("Statistiek bot "+ d.naam);
            w.println(">>**************************************************<<");
            w.println("I fire bullet amount       : "+ d.getSchotenBvh() );
            w.println("My hit bullet amount       : "+ d.getTreffersBvh() );
            w.println("                 Hit ratio : "+100.0d*d.getTreffersBvh()/d.getSchotenBvh()+"%");
            w.println("+-------------------------------------+");
            w.println("Enemy fire bullet amount   : "+ d.getSchoten() );
            w.println("Enemy  hit bullet amount   : "+ d.getTreffers() );
            w.println("                 Hit ratio : "+100.0d*d.getTreffers()/d.getSchoten()+"%");
            w.println("+-------------------------------------+");
            w.println("Gemiddelde afstand         : "+d.afstandGem);

            w.println("+-------------------------------------+");
*/
// guessFactors GF-kanon: [][1][0][2] : neutrale positie, constante snelheid, max. snelheid
// int[AFSTANDSEGMENTEN][POSITIESEGMENTEN][VERSNELLINGSSEGMENTEN][SNELHEIDSSEGMENTEN][2*GF_MIDDEN+1];
            for (int i = 0; i < AFSTANDSEGMENTEN; i++) {
               tekst = "waveGuessFactors["+i+"]";
               for (int j = 0; j < (2*GF_MIDDEN+1); j++) {
                  tekst += ", " + d.guessFactors[i][0][0][2][j];
               }
               log(tekst);
//               w.println(tekst);
            }
//            w.println("+-------------------------------------+");
// guessFactors GF-kanon: [][1][0][2] : neutrale positie, constante snelheid, max. snelheid
// int[AFSTANDSEGMENTEN][POSITIESEGMENTEN][VERSNELLINGSSEGMENTEN][SNELHEIDSSEGMENTEN][2*GF_MIDDEN+1];
            for (int i = 0; i < AFSTANDSEGMENTEN; i++) {
               tekst = "waveGuessFactors["+i+"]";
               for (int j = 0; j < (2*GF_MIDDEN+1); j++) {
                  tekst += ", " + d.guessFactors[i][1][0][2][j];
               }
               log(tekst);
//               w.println(tekst);
            }
//            w.println("+-------------------------------------+");
// guessFactors GF-kanon: [][1][0][2] : neutrale positie, constante snelheid, max. snelheid
// int[AFSTANDSEGMENTEN][POSITIESEGMENTEN][VERSNELLINGSSEGMENTEN][SNELHEIDSSEGMENTEN][2*GF_MIDDEN+1];
            for (int i = 0; i < AFSTANDSEGMENTEN; i++) {
               tekst = "waveGuessFactors["+i+"]";
               for (int j = 0; j < (2*GF_MIDDEN+1); j++) {
                  tekst += ", " + d.guessFactors[i][2][0][2][j];
               }
               log(tekst);
//               w.println(tekst);
            }
//            w.println("+-------------------------------------+");
// guessFactors GF-kanon: [][1][0][2] : neutrale positie, constante snelheid, max. snelheid
// int[AFSTANDSEGMENTEN][POSITIESEGMENTEN][VERSNELLINGSSEGMENTEN][SNELHEIDSSEGMENTEN][2*GF_MIDDEN+1];
            for (int i = 0; i < AFSTANDSEGMENTEN; i++) {
               tekst = "waveGuessFactors["+i+"]";
               for (int j = 0; j < (2*GF_MIDDEN+1); j++) {
                  tekst += ", " + d.guessFactors[i][0][0][0][j];
               }
               log(tekst);
//               w.println(tekst);
            }
//            w.println("+-------------------------------------+");
// guessFactors GF-kanon: [][1][0][2] : neutrale positie, constante snelheid, max. snelheid
// int[AFSTANDSEGMENTEN][POSITIESEGMENTEN][VERSNELLINGSSEGMENTEN][SNELHEIDSSEGMENTEN][2*GF_MIDDEN+1];
            for (int i = 0; i < AFSTANDSEGMENTEN; i++) {
               tekst = "waveGuessFactors["+i+"]";
               for (int j = 0; j < (2*GF_MIDDEN+1); j++) {
                  tekst += ", " + d.guessFactors[i][1][0][1][j];
               }
               log(tekst);
//               w.println(tekst);
            }
//            w.println("+-------------------------------------+");
// guessFactors GF-kanon: [][1][0][2] : neutrale positie, constante snelheid, max. snelheid
// int[AFSTANDSEGMENTEN][POSITIESEGMENTEN][VERSNELLINGSSEGMENTEN][SNELHEIDSSEGMENTEN][2*GF_MIDDEN+1];
            for (int i = 0; i < AFSTANDSEGMENTEN; i++) {
               tekst = "waveGuessFactors["+i+"]";
               for (int j = 0; j < (2*GF_MIDDEN+1); j++) {
                  tekst += ", " + d.guessFactors[i][2][0][2][j];
               }
               log(tekst);
//               w.println(tekst);
            }
//            w.println("+-------------------------------------+");
// guessFactors GF-kanon: [][1][0][2] : neutrale positie, constante snelheid, max. snelheid
// int[AFSTANDSEGMENTEN][POSITIESEGMENTEN][VERSNELLINGSSEGMENTEN][SNELHEIDSSEGMENTEN][2*GF_MIDDEN+1];
            for (int i = 0; i < AFSTANDSEGMENTEN; i++) {
               tekst = "waveGuessFactors["+i+"]";
               for (int j = 0; j < (2*GF_MIDDEN+1); j++) {
                  tekst += ", " + d.guessFactors[i][1][0][2][j];
               }
               log(tekst);
//               w.println(tekst);
            }
//            w.println("+-------------------------------------+");
// guessFactors GF-kanon: [][1][1][2] : neutrale positie, constante snelheid, max. snelheid
// int[AFSTANDSEGMENTEN][POSITIESEGMENTEN][VERSNELLINGSSEGMENTEN][SNELHEIDSSEGMENTEN][2*GF_MIDDEN+1];
            for (int i = 0; i < AFSTANDSEGMENTEN; i++) {
               tekst = "waveGuessFactors["+i+"]";
               for (int j = 0; j < (2*GF_MIDDEN+1); j++) {
                  tekst += ", " + d.guessFactors[i][1][1][2][j];
               }
               log(tekst);
//               w.println(tekst);
            }
//            w.println("+-------------------------------------+");
// guessFactors GF-kanon: [][1][2][2] : neutrale positie, constante snelheid, max. snelheid
// int[AFSTANDSEGMENTEN][POSITIESEGMENTEN][VERSNELLINGSSEGMENTEN][SNELHEIDSSEGMENTEN][2*GF_MIDDEN+1];
            for (int i = 0; i < AFSTANDSEGMENTEN; i++) {
               tekst = "waveGuessFactors["+i+"]";
               for (int j = 0; j < (2*GF_MIDDEN+1); j++) {
                  tekst += ", " + d.guessFactors[i][1][2][2][j];
               }
               log(tekst);
//               w.println(tekst);
            }

//            if (w.checkError()) {
//               log("Probleem bij wegschrijven statistiek!");
//            }
//            w.close();
            log(d.naam+"'s data opgeslagen");
         } catch (Exception e) {
            log("saveSatistiek(): IOException bij opslaan data " + d.naam + " : " + e);
         }
      }
   }

/**********************************************************************
** methods voor lezen/schrijven data:
**********************************************************************/
   private void inlezenDoelData(String naam, Doel d) {
log("inlezenDoelData(String naam="+naam+", Doel d)");

      try {
         ObjectInputStream in = new ObjectInputStream(
                                   new GZIPInputStream(
                                      new FileInputStream(getDataFile(HulpRoutines.getFilenaam(naam)))));
         d.guessFactors = (int[][][][][])in.readObject();
         in.close();
         log("data ingelezen: wordt nu getoond...");
         
         try {
            String tekst;
//            PrintStream w = new PrintStream(new RobocodeFileOutputStream(getDataFile(naam+".gf")));
//            w.println("+-------------------------------------+");
// guessFactors GF-kanon: [][1][0][2] : neutrale positie, constante snelheid, max. snelheid
// int[AFSTANDSEGMENTEN][POSITIESEGMENTEN][VERSNELLINGSSEGMENTEN][SNELHEIDSSEGMENTEN][2*GF_MIDDEN+1];
            for (int i = 0; i < AFSTANDSEGMENTEN; i++) {
               tekst = "waveGuessFactors["+i+"]";
               for (int j = 0; j < (2*GF_MIDDEN+1); j++) {
                  tekst += ", " + d.guessFactors[i][0][0][0][j];
               }
               log(tekst);
//               w.println(tekst);
            }
//            w.println("+-------------------------------------+");
// guessFactors GF-kanon: [][1][0][2] : neutrale positie, constante snelheid, max. snelheid
// int[AFSTANDSEGMENTEN][POSITIESEGMENTEN][VERSNELLINGSSEGMENTEN][SNELHEIDSSEGMENTEN][2*GF_MIDDEN+1];
            for (int i = 0; i < AFSTANDSEGMENTEN; i++) {
               tekst = "waveGuessFactors["+i+"]";
               for (int j = 0; j < (2*GF_MIDDEN+1); j++) {
                  tekst += ", " + d.guessFactors[i][1][0][0][j];
               }
               log(tekst);
//               w.println(tekst);
            }
//            w.println("+-------------------------------------+");
// guessFactors GF-kanon: [][1][0][2] : neutrale positie, constante snelheid, max. snelheid
// int[AFSTANDSEGMENTEN][POSITIESEGMENTEN][VERSNELLINGSSEGMENTEN][SNELHEIDSSEGMENTEN][2*GF_MIDDEN+1];
            for (int i = 0; i < AFSTANDSEGMENTEN; i++) {
               tekst = "waveGuessFactors["+i+"]";
               for (int j = 0; j < (2*GF_MIDDEN+1); j++) {
                  tekst += ", " + d.guessFactors[i][2][0][0][j];
               }
               log(tekst);
//               w.println(tekst);
            }
//            w.println("+-------------------------------------+");
// guessFactors GF-kanon: [][1][0][2] : neutrale positie, constante snelheid, max. snelheid
// int[AFSTANDSEGMENTEN][POSITIESEGMENTEN][VERSNELLINGSSEGMENTEN][SNELHEIDSSEGMENTEN][2*GF_MIDDEN+1];
            for (int i = 0; i < AFSTANDSEGMENTEN; i++) {
               tekst = "waveGuessFactors["+i+"]";
               for (int j = 0; j < (2*GF_MIDDEN+1); j++) {
                  tekst += ", " + d.guessFactors[i][0][1][0][j];
               }
               log(tekst);
//               w.println(tekst);
            }
//            w.println("+-------------------------------------+");
// guessFactors GF-kanon: [][1][0][2] : neutrale positie, constante snelheid, max. snelheid
// int[AFSTANDSEGMENTEN][POSITIESEGMENTEN][VERSNELLINGSSEGMENTEN][SNELHEIDSSEGMENTEN][2*GF_MIDDEN+1];
            for (int i = 0; i < AFSTANDSEGMENTEN; i++) {
               tekst = "waveGuessFactors["+i+"]";
               for (int j = 0; j < (2*GF_MIDDEN+1); j++) {
                  tekst += ", " + d.guessFactors[i][0][2][0][j];
               }
               log(tekst);
//               w.println(tekst);
            }
//            w.println("+-------------------------------------+");
// guessFactors GF-kanon: [][1][0][2] : neutrale positie, constante snelheid, max. snelheid
// int[AFSTANDSEGMENTEN][POSITIESEGMENTEN][VERSNELLINGSSEGMENTEN][SNELHEIDSSEGMENTEN][2*GF_MIDDEN+1];
            for (int i = 0; i < AFSTANDSEGMENTEN; i++) {
               tekst = "waveGuessFactors["+i+"]";
               for (int j = 0; j < (2*GF_MIDDEN+1); j++) {
                  tekst += ", " + d.guessFactors[i][0][0][1][j];
               }
               log(tekst);
//               w.println(tekst);
            }
//            w.println("+-------------------------------------+");
// guessFactors GF-kanon: [][1][0][2] : neutrale positie, constante snelheid, max. snelheid
// int[AFSTANDSEGMENTEN][POSITIESEGMENTEN][VERSNELLINGSSEGMENTEN][SNELHEIDSSEGMENTEN][2*GF_MIDDEN+1];
            for (int i = 0; i < AFSTANDSEGMENTEN; i++) {
               tekst = "waveGuessFactors["+i+"]";
               for (int j = 0; j < (2*GF_MIDDEN+1); j++) {
                  tekst += ", " + d.guessFactors[i][0][0][2][j];
               }
               log(tekst);
//               w.println(tekst);
            }
//            w.println("+-------------------------------------+");
// guessFactors GF-kanon: [][1][1][2] : neutrale positie, constante snelheid, max. snelheid
// int[AFSTANDSEGMENTEN][POSITIESEGMENTEN][VERSNELLINGSSEGMENTEN][SNELHEIDSSEGMENTEN][2*GF_MIDDEN+1];
            for (int i = 0; i < AFSTANDSEGMENTEN; i++) {
               tekst = "waveGuessFactors["+i+"]";
               for (int j = 0; j < (2*GF_MIDDEN+1); j++) {
                  tekst += ", " + d.guessFactors[i][1][1][2][j];
               }
               log(tekst);
//               w.println(tekst);
            }
//            w.println("+-------------------------------------+");
// guessFactors GF-kanon: [][1][2][2] : neutrale positie, constante snelheid, max. snelheid
// int[AFSTANDSEGMENTEN][POSITIESEGMENTEN][VERSNELLINGSSEGMENTEN][SNELHEIDSSEGMENTEN][2*GF_MIDDEN+1];
            for (int i = 0; i < AFSTANDSEGMENTEN; i++) {
               tekst = "waveGuessFactors["+i+"]";
               for (int j = 0; j < (2*GF_MIDDEN+1); j++) {
                  tekst += ", " + d.guessFactors[i][1][1][1][j];
               }
               log(tekst);
//               w.println(tekst);
            }

//            if (w.checkError()) {
//               log("Probleem bij wegschrijven ingelezen statistiek!");
//            }
//            w.close();
            log(naam+"'s ingelezen data opgeslagen in *.gf file");
         } catch (Exception e) {
            log("inlezen(): IOException bij tonen array data " + naam + " : " + e);
         }
      }
      catch (Exception ex) {
         log("GF data-file bestaat niet: GF initialiseren");
         // wall, accl, latv, distance, guessfactors
         d.guessFactors = new int[AFSTANDSEGMENTEN][POSITIESEGMENTEN][VERSNELLINGSSEGMENTEN][SNELHEIDSSEGMENTEN][2*GF_MIDDEN+1];
      }
   }


   private void opslaanDoelData(Doel d) {
      boolean dataruimte_beschikbaar = true;
      File    directory       = null;
      File[]  robotFiles      = null;
      long    totalBytes      = 0;
log("opslaanDoelData(doel = "+d.naam+" )");

      /* the code to check if there is enough room to save my data is based on
       * code from tzu.TheArtOfWar, by Ray Vermette, including the comments.
       *
       * If the data directory does not exist, yet, then
       * it can't possibly be full.
       */
      directory  = this.getDataDirectory();
      if (directory != null) {
          /*
           * If there are no files in the directory, then
           * we certainly have room.
           */
          robotFiles = directory.listFiles();
          if (robotFiles != null) {
             for (int i = 0; i < robotFiles.length; i++) {
                totalBytes += robotFiles[i].length();
             }
             if (totalBytes + (12500) <= MAX_DATA_DIRECTORY)
                dataruimte_beschikbaar = true;
             else
                dataruimte_beschikbaar = false;
          } // einde controle op aantal datafiles
      } // einde controle of directory bestaat

      if (dataruimte_beschikbaar) {
// wegschrijven data (pm 5kb per bot)
         try  {
            ObjectOutputStream out = new ObjectOutputStream(
                                        new GZIPOutputStream(
                                           new RobocodeFileOutputStream(getDataFile(HulpRoutines.getFilenaam(d.naam)))));
            out.writeObject(d.guessFactors);
            out.close();
         }
         catch (IOException ex) {
         }
      } // einde controle of voldoende ruimte aanwezig is.
   }


// log method: voor debugging ...
   public void log(String tekst) {
      if (DEBUGMODE > 0) out.println(tekst);
   }
/**********************************************************************
** Constructor:
**********************************************************************/
   public Jormungand() {
      oudePositie   = new Point2D.Double(0,0);
      huidigePositie= new Point2D.Double(0,0);
      nieuwePositie = new Point2D.Double(0,0);
      nieuweRichting= 0;
      voorwaards    = 1;

      minKogelsnelheid = 20D - 3D * MAXVUURKRACHT;
      maxKogelsnelheid = 20D - 3D * MINVUURKRACHT;

// initieer doel & kogel hashtabel:
      doel          = new Doel();
      kogels        = new Hashtable();

//      narc          = new Narcissus();
   }
/**********************************************************************
** Definitie van variabelen
**********************************************************************/
// overige objecten:
   private boolean adjustGun  = true; // kanon draait in tegengestelde richting als bot: blijft daardoor oorspronkelijke richting behouden
   private boolean adjustRadar= true; // radar draait in tegengestelde richting als bot: blijft daardoor oorspronkelijke richting behouden
   private int     beurtOvergeslagen;

// variabelen voor sturing van beweging en radar:
   private Slagveld  slagveld;

// variabelen voor beweging:
   private int              voorwaards;
   private double           nieuweRichting;
   public  Point2D.Double   nieuwePositie, huidigePositie, oudePositie;
   private boolean          uitwijken;
   public  static double    vGem;
           Random           random = new Random();

// variabelen voor kanon:
   private double  minKogelsnelheid;
   private double  maxKogelsnelheid;

// variabelen voor uitwijkingsbeleid:
   public  static  Hashtable doelen = new Hashtable(); // hier initieren omdat in constructor hashtable iedere keer opnieuw wordt geinitieerd en opgedane data verloren gaat.

// variabelen voor overleving ;-)
   public          Hashtable kogels;

// variabelen die huidig doel of directe bedreiging bevatten:
   public  boolean   meleeModus;
   private static Doel      doel;

// grapher variabelen
//   Narcissus       narc;

/*
// GL4Java variabelen...
   GLRenderer        renderer;
   PointGL           destPoint = new PointGL();
   LabelGL           destLabel = new LabelGL("");
   PointGL[]         potPoint  = {new PointGL(), new PointGL(), new PointGL(), new PointGL()
                                 ,new PointGL(), new PointGL(), new PointGL(), new PointGL()};
   LineGL            kogelLine = new LineGL(),
                     kogelHoek = new LineGL();
   EllipseGL         kogelGrens= new EllipseGL();
   EllipseGL[]       kGrens    = {new EllipseGL(), new EllipseGL(), new EllipseGL()
                                 ,new EllipseGL(), new EllipseGL(), new EllipseGL()};
   LineGL[]          kLine     = {new LineGL(), new LineGL(), new LineGL()
                                 ,new LineGL(), new LineGL(), new LineGL()};
*/

/**********************************************************************
** innerclass met de golf.
**********************************************************************/
class MicroGolf extends Condition {

   Point2D.Double   startPositie;
   double 		     richting, afstand, cirkelRichting, kogelSnelheid;
   int[] 			  stats;

   public boolean test()
   {
      int index;

      if ((Jormungand.doel.positie).distance(startPositie) <= (afstand+=kogelSnelheid)){
         try {
            index = (int)Math.round(Utils.normalRelativeAngle(HulpRoutines.bepaalRichting( Jormungand.doel.positie
                                                                                         , startPositie
                                                                                         ) - richting
                                                             )/cirkelRichting
                                   ) + GF_MIDDEN; 
//            System.out.println( "test(): index = " + index + ", doelPositie=" + Jormungand.doel.positie);
            stats[index]++;
         } catch (ArrayIndexOutOfBoundsException e){
//            System.out.println( "test(): ArrayIndexOutOfBoundsException " + (int)(Math.round(Utils.normalRelativeAngle(HulpRoutines.bepaalRichting(startPositie, Jormungand.doel.positie) - richting)/cirkelRichting) + GF_MIDDEN) );
         }
         removeCustomEvent(this);
      }
      return false;
   }
}
/**********************************************************************
** einde innerclass met de golf.
**********************************************************************/
}
/**********************************************************************
** EINDE CLASS Jormungand
**********************************************************************/